import logging
import threading
import time
from typing import Any, Dict

from celery import Celery
from celery.events import EventReceiver
from celery.events.state import State

from celery_farmer.statsd import StatsClient
from celery_farmer.task import get_tags

logger = logging.getLogger(__name__)


class EventListener(threading.Thread):
    def __init__(self, celery_app: Celery, statsd_client: StatsClient) -> None:
        super().__init__()
        self.daemon = True

        self.celery_app = celery_app
        self.state = self.celery_app.events.State()
        self.statsd_client = statsd_client

        self.is_terminated = False

        self.timings: Dict[str, Dict[str, Any]] = {}

    def stop(self) -> None:
        logger.info('Stopping EventListener')
        self.is_terminated = True

    def run(self) -> None:
        logger.info(f'Running EventListener with daemon: {self.isDaemon()}')
        sleep_time = 1
        while not self.is_terminated:
            try:
                if sleep_time < 10:
                    sleep_time *= 2

                with self.celery_app.connection() as connection:
                    receiver = EventReceiver(
                        connection,
                        handlers={'*': self.on_event},
                        app=self.celery_app
                    )
                    sleep_time = 1
                    receiver.capture(limit=None, timeout=None, wakeup=True)
            except Exception as e:
                logger.debug(e, exc_info=True)
                logger.error(f'Failed to capture events: {e}')
                time.sleep(sleep_time)

    def on_event(self, event: Dict[str, Any]) -> None:
        if event['type'].startswith('task-'):
            logger.debug(f'Got event {event}')

            self.state.event(event)
            task = self.state.tasks.get(event['uuid'])

            self.track_event(task)
            self.track_timing(task)

    def track_event(self, task: State.Task) -> None:
        task_tags = get_tags(task)
        self.statsd_client.incr(f'tasks.counts.{task.type}', tags=task_tags)

    def track_timing(self, task: State.Task) -> None:
        now = time.time()
        task_timings = self.timings.get(task.uuid, {})
        task_tags = get_tags(task)

        if task.type == 'task-received':
            task_timings['received'] = now
        elif task.type == 'task-started':
            task_timings['started'] = now

            time_received = task_timings.get('received')
            if time_received:
                time_not_started = now - time_received
                self.statsd_client.timing(
                    'tasks.times.not_started',
                    time_not_started * 1000,
                    tags=task_tags
                )
            else:
                logger.error(f"Task {task.uuid} didn't have received time")
        elif task.type == 'task-succeeded' or task.type == 'task-failed':
            time_started = task_timings.get('started')
            if time_started:
                execution_time = now - time_started
                self.statsd_client.timing(
                    'tasks.times.execution',
                    execution_time * 1000,
                    tags=task_tags
                )
            else:
                logger.error(f"Task {task.uuid} didn't have started time")

            task_timings = {}

        if len(task_timings) > 0:
            self.timings[task.uuid] = task_timings
        elif self.timings.get(task.uuid):
            del self.timings[task.uuid]



from itertools import chain
import logging
import threading
import time
from typing import Set

from celery import Celery

from celery_farmer.broker import RedisBroker
from celery_farmer.statsd import StatsClient

logger = logging.getLogger(__name__)


class QueueLengths(threading.Thread):
    def __init__(self, celery_app: Celery, statsd_client: StatsClient,
                 poll_time: float = 10.0) -> None:
        super().__init__()

        self.celery_app = celery_app
        self.statsd_client = statsd_client
        self.poll_time = poll_time

        self.is_terminated = False

        self.broker = RedisBroker(
            redis_uri=self.celery_app.broker_connection().as_uri())

    def stop(self) -> None:
        logger.info('Stopping QueueLengths')
        self.is_terminated = True

    def run(self) -> None:
        logger.info('Running QueueLengths')
        while not self.is_terminated:
            logger.debug('Sending heart beat')
            self.statsd_client.incr('heartbeats.queue_lengths')
            try:
                queues = self._get_active_queues()
                for queue_name in queues:
                    self._track_queue_length(queue_name)
            except Exception as e:
                logger.error(f'Failed to track queue lengths: {e}')
            finally:
                poll_time = self.poll_time
                time.sleep(poll_time)

    def _get_active_queues(self) -> Set[str]:
        response = self.celery_app.control.inspect().active_queues()
        if response is None:
            return set()

        queues = [
            [queue['name'] for queue in node]
            for node in response.values()
        ]

        return set(chain(*queues))

    def _track_queue_length(self, queue_name: str) -> None:
        length = self.broker.get_queue_length(queue_name)
        tags = {'queue': queue_name}

        self.statsd_client.gauge('queues.length', length, tags=tags)



import redis


class Broker:
    def get_queue_length(self, queue: str) -> int:
        raise NotImplementedError(
            'Subclass needs to implement get_queue_length')


class RedisBroker(Broker):
    def __init__(self, redis_uri: str) -> None:
        self.client = redis.StrictRedis.from_url(redis_uri)

    def get_queue_length(self, queue: str) -> int:
        return self.client.llen(queue)



import logging
from typing import Any, Dict, Optional

import statsd

logger = logging.getLogger(__name__)


class StatsClient:

    def __init__(self, config: Dict[str, Any]) -> None:
        host = config.get('host', 'localhost')
        port = config.get('port', 8125)
        prefix = config.get('prefix', 'farmer')

        client = statsd.StatsClient(host, port, prefix=prefix)
        logger.info(
            f'Configuring StatsD, host: {host}, port: {port}, prefix: {prefix}'
        )
        self.client = client

    def timer(self, stat: str, tags: Optional[Dict[str, str]] = None,
              rate: int = 1) -> None:
        self.client.timer(self._get_stat(stat, tags), rate)

    def timing(self, stat: str, delta: float,
               tags: Optional[Dict[str, str]] = None, rate: int = 1) -> None:
        self.client.timing(self._get_stat(stat, tags), delta, rate)

    def incr(self, stat: str, count: int = 1,
             tags: Optional[Dict[str, str]] = None, rate: int = 1) -> None:
        self.client.incr(self._get_stat(stat, tags), count, rate)

    def decr(self, stat: str, count: int = 1,
             tags: Optional[Dict[str, str]] = None, rate: int = 1) -> None:
        self.client.decr(self._get_stat(stat, tags), count, rate)

    def gauge(self, stat: str, value: float, rate: int = 1,
              tags: Optional[Dict[str, str]] = None, delta: bool = False
              ) -> None:
        self.client.gauge(self._get_stat(stat, tags), value, rate, delta)

    def set(self, stat: str, value: float,
            tags: Optional[Dict[str, str]] = None, rate: int = 1) -> None:
        self.client.set(self._get_stat(stat, tags), value, rate)

    def _get_stat(self, stat: str, tags: Optional[Dict[str, str]]) -> str:
        if tags is not None:
            stat = f'{stat},{self._format_tags(tags)}'
        return stat

    def _format_tags(self, tags: Dict[str, str]) -> str:
        tag_pairs = (f'{k}={v}' for k, v in tags.items())
        return ','.join(tag_pairs)



import logging
from typing import Any, Dict

from celery import Celery

from celery_farmer.enable_events import EnableEvents
from celery_farmer.event_listener import EventListener
from celery_farmer.queue_lengths import QueueLengths
from celery_farmer.statsd import StatsClient

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)s %(message)s'
)

logger = logging.getLogger(__name__)


class Farmer:
    def __init__(self, broker_url: str, poll_time: float,
                 statsd_config: Dict[str, Any]) -> None:
        self.broker_url = broker_url
        self.celery_app = Celery(broker=self.broker_url)

        self.statsd_client = StatsClient(statsd_config)

        self.enable_events_thread = EnableEvents(self.celery_app, poll_time)
        self.queue_lenghts_thread = QueueLengths(self.celery_app,
                                                 self.statsd_client, poll_time)

        self.event_listener_thread = EventListener(self.celery_app,
                                                   self.statsd_client)

    def start(self) -> None:
        logger.info(f'Starting Farmer with broker {self.broker_url}')
        self.enable_events_thread.start()
        self.queue_lenghts_thread.start()

        self.event_listener_thread.start()

    def stop(self) -> None:
        logger.info('Stopping farmer')
        self.event_listener_thread.stop()
        self.enable_events_thread.stop()
        self.queue_lenghts_thread.stop()



import logging
import threading
import time

from celery import Celery

logger = logging.getLogger(__name__)


class EnableEvents(threading.Thread):
    def __init__(self, celery_app: Celery, poll_time: float = 10.0) -> None:
        super().__init__()

        self.celery_app = celery_app
        self.poll_time = poll_time

        self.is_terminated = False

    def stop(self) -> None:
        logger.info('Stopping EnableEvents')
        self.is_terminated = True

    def run(self) -> None:
        logger.info('Running EnableEvents')
        while not self.is_terminated:
            logger.debug('Enabling events')
            try:
                self.celery_app.control.enable_events()
            finally:
                poll_time = self.poll_time
                time.sleep(poll_time)



#!/usr/bin/env python
import setuptools

with open('README.md', 'r') as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name='celery-farmer',
    version='0.1.0',
    author='Wolt Enterprises Oy',
    author_email='support@wolt.com',
    description='Farmer will monitor how the Celery cluster is behaving',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/woltapp/celery-farmer',
    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
    ],
    install_requires=[
        'statsd>=3.2.0,<4.0',
        'redis>=2.10.0,<4.0',
        'celery>=4.0,<5.0',
        'click>=6.7,<8.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    entry_points=dict(
        console_scripts='celery-farmer = celery_farmer.cli:cli'
    ),
)



from functools import partial
from inspect import FrameInfo
import signal
from typing import Any, Dict

import click

import celery_farmer
from celery_farmer.application import Farmer


@click.group()
@click.version_option(celery_farmer.__version__)
def cli() -> None:
    pass


@cli.command()
@click.option('--broker', '-b', envvar='BROKER', required=True,
              help="Celery app's broker url")
@click.option('--poll-time', envvar='POLL_TIME', type=float, default=10,
              help='Specify polling time')
@click.option('--statsd-host', '-sh', envvar='STASTD_HOST', help='Statsd host')
@click.option('--statsd-port', '-sp', envvar='STATSD_PORT', type=int,
              help='Statsd port')
@click.option('--statsd-prefix', '-spr', envvar='STATSD_PREFIX',
              help='Statsd prefix')
def start(broker: str, poll_time: float, statsd_host: str, statsd_port: int,
          statsd_prefix: str) -> None:
    def stop_farmer(farmer: Farmer, signal: int, frame: FrameInfo) -> None:
        farmer.stop()

    def construct_statsd_configs(host: str, port: int, prefix: str
                                 ) -> Dict[str, Any]:
        config: Dict[str, Any] = {}
        if host:
            config['host'] = host
        if port:
            config['port'] = port
        if prefix:
            config['prefix'] = prefix
        return config

    farmer = Farmer(
        broker,
        poll_time,
        construct_statsd_configs(statsd_host, statsd_port, statsd_prefix)
    )
    farmer.start()

    signal.signal(signal.SIGINT, partial(stop_farmer, farmer))
    signal.pause()


if __name__ == '__main__':
    cli()
